/**
 * SyncManager:
 * 
 * Implements 2-way sync of objects between app and server
 *
 * @author zhoujun@waveo.com & young@waveo.com
 */
Ti.include('/globals.js');

var SyncManager = App.Class(App.Controller.WOManager, {
	className	: 'SyncManager',
	alias		: 'sync',

	EVENT		: {
		//--- handles these events:
		doSyncObjects		: 'sync.syncObjects',
		doSyncImages		: 'sync.syncImages',
		deviceTokenUpdated	: 'notify.deviceTokenUpdated',	// MUST be the same as NotifyMgr event
		//--- produces these events:
		ready				: 'sync.ready',
		//---  event sub-types, used with an Object Type prefix, eg. WOCardMaster.sync.done
		syncDone			: 'sync.done',		// when local object sync'd Up to server successfully, also
                                        		// used alone (without prefix) when call to syncObjects() is completed
		noUpdate			: 'sync.noUpdate',	// when local object sync'd Up but no update from server
		syncUpdate          : 'sync.update',    // when local object sync'd Up and get update from server
    },

	//---  Constructor  ---------------        
	initialize : function(param){
		var self = this;
		self.base(param);   

		self.inited 		= false;
		
		self.syncTypesAlias    = {// Model.alias : Model.className
		    settings	: 'WOSettings',
            cardins		: 'WOCard',
            cardmas		: 'WOCardMaster',
            // cardnum    : 'WOCardNumber',
            rwdsetins  : 'WORewardset',
            rwdsetmas  : 'WORewardsetMaster',
            offerins   : 'WOOffer',
            offermas   : 'WOOfferMaster',
            usage		: 'WOUsage',
            watchdog	: 'WOLog',
            message     : 'WOMessage',
            // location	: 'WOLocation',
            points		: 'WOPoints',
            // pointsSt 	: 'WOPointsState',
            perkcode	: 'WOPerkCode',
        },
    
        self.userSyncUpTypes = [
            'WOSettings',
            'WOCard',
            'WORewardset',
            'WOOffer',
            'WOUsage',
            'WOLog',
            'WOMessage',
            'WOPoints',
            // 'WOCardMaster',
        ],
		
		self.imageQueue = [];				// array of image URLs to be pre-fetched by syncImage()
		
        self.AllSyncUpTypes = [];
        for(objectAlias in self.syncTypesAlias)
        	self.AllSyncUpTypes.push(self.syncTypesAlias[objectAlias]);
        
		if(!self.inited){
			var imagePath = Titanium.Filesystem.getFile(Titanium.Filesystem.applicationDataDirectory, "images");  //TODO [ZJ], need to move to bootstrap or initAppliation?
			if(!imagePath.exists()) imagePath.createDirectory();
			
			if(! App.Config.IS_ANDROID) imagePath.remoteBackup = false; //For icloud backup policy ... please remember to add these lines

			self.inited = true;
		}
		
		self.listenForEvents();
        Ti.App.fireEvent(self.EVENT.ready);		//------ We're Ready! -------
	},

	/**
	 * hook up handlers for Events that we handle, both Internal & Extneral (ie. service requests)
	 */
	listenForEvents : function(){
		var self = this;

		//----- Internal housekeeping events (set these up before External) -----
		
		//  remove local records after sync of Usage & Log objects
		// Ti.App.addEventListener(App.Const.MODEL_USAGE+App.Event.SYNC_DONE,  self._postUsageSync); //removed by ZJ, no need
        // Ti.App.addEventListener(App.Const.MODEL_LOG+App.Event.SYNC_DONE,  self._postLogSync);  //removed by ZJ, no need

        //------ External Events  ------------------------
		//--- doSyncObjects ---
		Ti.App.addEventListener(self.EVENT.doSyncObjects, function(evt){
			if('undefined' != typeof evt.objTypes){
				self.syncObjects(evt.objTypes);
			}else{
				self.syncObjects();
			}
		});
		Ti.App.addEventListener(self.EVENT.doSyncImages, function(evt){
			self.syncImages();
		});
		
		//--- ready (App) ------
		Ti.App.addEventListener($.app.EVENT.ready, function(evt){
			self.startUpSync(evt);
		});
        //--- deviceTokenUpdated (notifyMgr) -----
        Ti.App.addEventListener(self.EVENT.deviceTokenUpdated, function(evt){
        	self.pushDeviceToken(evt);
       	});
	},

	/**
	 * Called by other Object Managers (eg. $.cardMgr) to request SyncMgr to automatically sync their objects
	 * when events (in eventList) occurs
	 * 
	 * Used to deal with post-object updates that requires IMMEDIATE SYNC with server
	 */
	registerObjectSync : function(eventList, objectClass){
		var self = this;

		//--- allow single (string) or array of events ---
		var _events = ('string' == typeof eventList) ? [eventList] : eventList;

		for(var i=0; i < _events.length; i++){
			Ti.App.addEventListener(_events[i], function(evt){
				self.syncObjects([objectClass]);
			});
		}	
	},

	//--------------------------   Event Handlers  ---------------------------------
	startUpSync : function(){
		var self = this;
		
        setTimeout(function(e){
            self.syncObjects();		// sync all objects
        }, 4000);
	},

	//--------------------------  Synchronization Services  ------------------------
  
     /**
     * Primary method for initiating sync of updated local objects with Server & vice versa
     * 
     * @param {array}      list of object types (listed in tag2ClassMap) to sync UP to server
     *                     when absent => sync ALL object types
     */ 
    syncObjects : function(syncUpTypes, pushMode, callback){
        var self = this;
        // JSON structure to send up to server

        //--- prepare local Object updates to SyncUp to server ---
        if(!syncUpTypes || syncUpTypes.length == 0) syncUpTypes = self.AllSyncUpTypes;       // default to all types when called with no parameters
        
        var syncUpObjects = {};
        var total = 0;
        for(var i=0; i < syncUpTypes.length; i++) {
            var className = syncUpTypes[i];
            if(!className) {
                Ti.API.error('SyncManager::syncObjects - invalid syncUp object type: '+className);
                continue;
            }
            var classObject = App.Helper.requireModel(className);
            var model = new classObject();
            var objectList = model.getSyncUp();
            if(pushMode && !objectList.length) continue;
            var lastSyncTime = self.getLastSyncTimestamp(className);
            var objectPackage = {
                type    : model.alias,
                stamp   : lastSyncTime,
                data    : objectList,
            }
            syncUpObjects[model.alias] = objectPackage;
            total += objectList.length;
        }
        // log ...
        Ti.API.error('-----+-----+-----+-----+-----+-----+-----+-----+-----+----- Sync start -----+-----+-----+-----+-----+-----+-----+-----+-----+-----');
        for(var objAlias in syncUpObjects) {
        	var className = self.syncTypesAlias[objAlias];
        	var lastSyncTime = syncUpObjects[objAlias].stamp;
        	var up = syncUpObjects[objAlias].data.length;
        	Ti.API.error('SyncManager::syncObjects - syncing '+className+': 				'+up+' up 				lastSyncTime: '+new Date(lastSyncTime*1000).toString('yyyy-MM-dd HH:mm:ss'));
        }
        
        //push mode, but nothing to push, so no need to call server, return directly.
        if(pushMode && !total) {
        	Ti.API.error('SyncManager::syncObjects - push mode, but nothing to push, no need to call server, return directly, sync cancelled.');
        	
        	if(callback) callback();
        	
        	return;
        }

        //---------------  Call Server  -------------
        var payload = {
			objects		: JSON.stringify(syncUpObjects),
            //--- extended payloads ---
            appInfo     : JSON.stringify($.configMgr.getAppInfo()),
            deviceInfo  : JSON.stringify($.configMgr.getDeviceInfo()),
        };
        var callParam = {
            action		: App.Action.SYNCOBJECTS,
            data		: payload,
            callback	: _syncCallback,
            errorback 	: _syncErrorback,
        };
        self.callServer(callParam);
        
		//------------------  Callback Handlers  ----------------       
        function _syncCallback(r) {
        	//------------- Handle Sync results ------------
            self.updateSyncStatus(r.result.objects);
            //------------- Handle Sync DOWN  ---------------
            self.updateSyncDownObjects(r.objects, true);
            self._logSyncResult(r);
			//------------- entire sync completed ---------------
            Ti.App.fireEvent(self.EVENT.syncDone);
            
            if(callback) callback();
        }
        
        function _syncErrorback(r) {
        	Ti.API.error('SyncManager::syncObjects - failed to sync, error:');
            Ti.API.error(r);
            // alert('Network Error (232)');
            Ti.App.fireEvent(self.EVENT.syncDone);
            
            if(callback) callback();
        }
    },
    
    /**
     * Update local copy of Objects structure received from Server
     * 
     * @param {object}      object->{ classTag1->{ type, stamp, data->[objects] }, classTag2->{}... }
     */
    updateSyncDownObjects : function(objects, updateSyncTime) {
        var self = this;
        var className = '';
        
        if('message' in objects) {//Patch for reorder objects.
        	var message = objects.message;
        	delete objects.message;
        	objects.message = message;
        }
        
        updateSyncTime =  updateSyncTime || false;

        //--- update each type/class of objects in turn 
        for(var objectAlias in objects) {
            //--- update timeStamp for this object type ---
            className = self.syncTypesAlias[objectAlias];
            if(className) {
                
                if(objects[objectAlias].data.length > 0) {
                    var classObject = App.Helper.requireModel(className);
                    var model = new classObject();
                    //--- update local objects ---
                    var numUpdated = model.saveSyncDown(objects[objectAlias].data);
                    // //--- update timeStamp for this object type ---
                    // Ti.API.info('SyncManager::syncObjects - synced down '+className+ ', total[down]: '+ numUpdated+', lastSyncTimestamp: '+lastSyncTimestamp);                   
                    //------  dispatch Post-Sync Events to trigger post-sync handlers  -------
		
		            // Note: event (App.Const.SYNC_UPDATE) for object types that have been updated
		            // locally with latest version from server, ie. sync'd down,
		            // have been dispatched within updateSyncDownObjects()                      
                    Ti.App.fireEvent(className+'.'+self.EVENT.syncUpdate, { count : numUpdated });
                
                } else {
                	// Dispatch 'No update' event (App.Const.SYNC_NOUPDATE) for object types that
                	// were sync'd up but got no update from server   
                	// Ti.API.info('SyncManager::syncObjects - no server updates for '+className);
                	Ti.App.fireEvent(className+'.'+self.EVENT.noUpdate);
                }
                // when all saved, update last sync time.
                if(updateSyncTime) {
	                var lastSyncTimestamp = objects[objectAlias].stamp;
	                self.setLastSyncTimestamp(className, lastSyncTimestamp);
                }
            }
            else {
            	
                Ti.API.error('SyncManager::updateSyncDownObjects - unknown object type skipped: ' + objectAlias);
            }
        }
    },
    
    updateSyncStatus : function(objects){
        var self = this;
        var className = '';
        //--- update each type/class of objects in turn 
        for(var objectAlias in objects) {
            //TODO [ZJ]
            //--- update timeStamp for this object type ---
            className = self.syncTypesAlias[objectAlias];
            if(className) {
                var classObject = App.Helper.requireModel(className);
                var model = new classObject();
                //--- update local objects ---
                var numUpdated = model.syncDone(objects[objectAlias]);
                // Ti.API.info('SyncManager::_updateSyncStatus - updated '+className+ ', total: '+ numUpdated);
            }
            else {
                Ti.API.error('SyncManager::_updateSyncStatus - unknown object type skipped: ' + objectAlias);
            }
        }
    },
    
    _logSyncResult : function(r){
    	var self = this;
    	Ti.API.error('-----+-----+-----+-----+-----+-----+-----+-----+-----+----- Sync Result -----+-----+-----+-----+-----+-----+-----+-----+-----+-----');
    	for(var objAlias in self.syncTypesAlias) {
    		var className = self.syncTypesAlias[objAlias];
    		var lastSyncTime = new Date(self.getLastSyncTimestamp(className)*1000);
    		if('undefined' == typeof r.result.objects[objAlias] && 'undefined' == typeof r.objects[objAlias]) continue;
    			// Ti.API.warn('SyncManager::syncObjects - no server updates for '+className+':				lastSyncTime: '+lastSyncTime.toString('yyyy-MM-dd HH:mm:ss'));
    			// continue;
    		// }
    		var up = 0, down = 0;
    		if('undefined' != typeof r.result.objects[objAlias]) for(var id in r.result.objects[objAlias]) up++;
    		if('undefined' != typeof r.objects[objAlias]) down = r.objects[objAlias].data.length;
    		Ti.API.error('SyncManager::syncObjects - synced '+className+':				'+up+' up '+down+' down'+'				newSyncTime: '+lastSyncTime.toString('yyyy-MM-dd HH:mm:ss'));
    	}
    	Ti.API.error('-----+-----+-----+-----+-----+-----+-----+-----+-----+----- Sync End -----+-----+-----+-----+-----+-----+-----+-----+-----+-----');
    },
    
    //for last sync timestamp 
    getLastSyncTimestamp : function(className) {
        var self = this;
        return Ti.App.Properties.getInt(className+'.lastSyncTimestamp', 0);
    },
    
    setLastSyncTimestamp : function(className, timestamp) {
        var self = this;
        if(timestamp >= 0) Ti.App.Properties.setInt(className+'.lastSyncTimestamp', timestamp);
    },
    
    resetSyncTimestamp : function() {
        var self = this;
        for(var objectAlias in self.syncTypesAlias) {
            self.setLastSyncTimestamp(self.syncTypesAlias[objectAlias], 0);
        }
    },
   
     /**
     * Safely Reset all local objects to server objects, except for Account Object
     * 
     * Used during app upgrades and for emergency recovery of app database
     */
    reset : function(){
    	
    },
    
    resetAll : function(){
    	var self = this;
    	//clear All local objects
    	for(var alias in self.syncTypesAlias) {
    		var className = self.syncTypesAlias[alias];
            var classObject = App.Helper.requireModel(className);
            if(!classObject) continue;
            var model = new classObject();
            model.reset();
            self.setLastSyncTimestamp(className, 0);
    	}
    	//clear app properties
    	// var props = Ti.App.Properties.listProperties();
		// for (var i=0, ilen=props.length; i<ilen; i++){
		    // var value = Ti.App.Properties.getString(props[i]);
		    // Ti.API.info(props[i] + ' = ' + value);
		// }
    	
    	return true;
    },

    /**
     * Reset all personalized data objects. Used mainly by Logout() of AuthMgr
     */ 
    resetUserObjects : function() {
        var self = this;
        for(var i=0; i < self.userSyncUpTypes.length; i++) {
            var className = self.userSyncUpTypes[i];
            var classObject = App.Helper.requireModel(className);
            var model = new classObject();
            model.reset();
            self.setLastSyncTimestamp(className, 0);
        }
        // self.resetAll();
        return true;
    },
    
    pushUserObjects : function(callback) {
        var self = this;
        self.syncObjects(self.userSyncUpTypes, true, callback);
    },
    
    //------------------  Sync of Non-Objects  ------------------

	/**
	 * Submit Device Token required for Push Notification
	 */
    pushDeviceToken : function(param){
        var self = this;

		if('undefined' != typeof param.deviceToken){
			var myDeviceToken = param.deviceToken;
		} else {
			var myDeviceToken = $.configMgr.getDeviceToken();
		}		
        if(! myDeviceToken || myDeviceToken=='') return;
		
        //-------------  Call Server  -------------
		var authState = $.authMgr.getAuthState();
        var payload = {
            deviceToken     : myDeviceToken,
            enabled			: 'Y',					// Young: hardcode for now, determine later if necc to provide this
            //--- extended payloads ---
            token			: authState.token,
            deviceInfo      : JSON.stringify($.configMgr.getDeviceInfo()),
        };
        var callParam = {
            action      : App.Action.UPDATEDEVTOKEN,
            data        : payload,
            errorback   : function(result){
            	Ti.API.error('SyncMgr::pushDeviceToken - ' + result);
            }
        };
        self.callServer(callParam);
    },

	pushLocationInfo : function(param){
		var self = this;

        //-------------  Call Server  -------------
		var authState = $.authMgr.getAuthState();
        var payload = {
            locationInfo	: JSON.stringify($.configMgr.getLocationInfo()),
            //--- extended payloads ---
            token			: authState.token,
            deviceInfo      : JSON.stringify($.configMgr.getDeviceInfo()),
        };
        var callParam = {
            action      : App.Action.UPDATELOCATION,
            data        : payload,
            errorback   : function(result){
            	Ti.API.error('SyncMgr::pushLocation - ' + result);
            }
        };
        self.callServer(callParam);
        
        $.watch.dog(self, 'pushLocationInfo', payload.locationInfo);
	},

    //------------  Sync'ing of image files  -----------

	/**
	 * Add image URLs to queue for sync later when syncImages() is called
     * @param {object}     Object attributes:
	 *      - urls         {array} of image file URLs to sync later
	 */
	queueImages : function(param){
		var self = this;
		self.imageQueue = self.imageQueue.concat(param.urls);
	},

	imagesQueued : function(){
		var self = this;
		return self.imageQueue.length;		
	},

     /**
     * Asynchronous fetching of image files and cache locally, tracked in 'CACHE' database
     * Cache NOT cleared during Logout
     * 
     * @param {object}     Object attributes:
     *      - pcallback    sync progress callback for display of progress bar, etc (optional)
     *      - scallback    sync completed callback  (optional)
     *      - maxProcess   max. no. of concurrent processes for image fetching allowed (optional)
     */    	
	syncImages : function(param){
	    var self = this;
	    
	    param = param || {};
	    
	    var taskPool = App.Helper.dedupe(self.imageQueue);
	    self.imageQueue = [];
	    
        var progressCallback    = param.pcallback;
        var successCallback     = param.scallback;       
        var maxProcess          = param.maxProcess || 5;
        
        //----- Task function to fetch an image -----
        var totalTasks      = taskPool.length;
        var processCount    = 0;
        var index           = 0;
        var doneCount       = 0;
        var taskCount       = 0;

        //---- Timer function to spawn INITIAL pool of Task processes  ----
        var spawnLimit = (totalTasks > maxProcess) ? maxProcess : totalTasks;
        var spawnCount = 0;
        var spawnInterval = 5;
        
        //  Let's GO !
        var spawner = setInterval(function(){
                if(spawnCount < spawnLimit){
                    spawnCount++;
                    Ti.API.debug('---- spawnTask: '+spawnCount);
                    _doTask();            
                } else {
                    clearInterval(spawner);
                }
            }, spawnInterval);

        Ti.API.debug('Started image sync...total='+totalTasks);
        
        //---------------------  Private Methods  -----------------------
        function _doTask(){
            processCount++;
            
            var taskID = taskCount++;
            Ti.API.debug('>>>> doTask start #'+taskID);
            
            if((processCount > maxProcess) || (taskCount > totalTasks)) {
                //--- there may be extra tasks spawn due to timing  ---
                processCount--;
                Ti.API.error('SyncManager::syncImage.doTask: skip extra tasks.  processCount='+processCount+'  taskCount='+taskCount+' of '+totalTasks);
                Ti.API.debug('<<<< doTask end #'+taskID);
                return;
            }
            self.fetchImage({
                url         : taskPool[taskID],
                callback    : _doImageFetched,
                errorback   : _doFetchFailed,
            });           
            Ti.API.debug('<<<< doTask end #'+taskID);
            
            //----------------------  Callback Handlers  ------------------------
            function _doImageFetched(path, theurl){
                //--- completed fetch of image successfully, update status ---
                processCount--;
                doneCount++;
                Ti.API.debug('>>>  fetchImage #'+taskID);
                
                //--- report progress to callback provided ---
                if(progressCallback)    progressCallback(doneCount, theurl);

                if(doneCount == totalTasks){
                    //--- report that we are done ---
                    Ti.API.info('SyncManager::syncImages - completed sync of '+totalTasks+' images');
                    if(successCallback)     successCallback(totalTasks);
                } else if(taskCount < totalTasks){
                    //--- spawn new process ---
                    setTimeout(_doTask, 5);
                }
            }
  			function _doFetchFailed(){
                processCount--;
                doneCount++;
                Ti.API.error('>>>  fetchImage #'+taskID+' failed');
                
                if(doneCount == totalTasks){
                    Ti.API.info('SyncManager::syncImages - completed sync of '+totalTasks+' images');
                    if(successCallback)     successCallback(totalTasks);
                } else if(taskCount < totalTasks){
                    //--- spawn new process ---
                    setTimeout(_doTask, 5);
                    if(progressCallback)    progressCallback(doneCount);
                }
            }
        }
	},
	
	fetchImage : function(param){
		var self = this;
        App.Helper.requireModel('WOImage');

        if(param.url==''){
            Ti.API.error('fetchImage: empty image URL!');		// FIXME to remove
            return param.url;
        }
		var image = new App.Model.WOImage(param.url);
		image.initialize();

		if(image.fetch() && image.data.filesize > 0) {
			//--- if image file already exist, just return it ---
			if(param.callback) param.callback(image.getFilepath(), param.url);
			return image.getFilepath();
		}else{
			Ti.API.error('Fetch image : '+param.url);
        	//--- Image not found locally, fetch it from remote  ---
	        App.Helper.fetchRemoteImage({
	        	url			: param.url,
	        	callback	: function(result){
	        		Ti.API.error('Fetch image success : '+param.url);
	        		if(param.callback)    param.callback(result.filepath, result.originalURL);
	        	},
	        	errorback	: function() {
	        		Ti.API.error('Fetch image failed : '+param.url);
	        		param.errorback();
	        	},
	        });
			return param.url;
		}
	}

});

module.exports = SyncManager;

//--- End of file: SyncManager.js
