/**
 * Base class of the main Application Controller
 * 
 * @author young@waveo.com
 */
Ti.include('/globals.js');
if (App.Config.IS_ANDROID) {
	Ti.include('/waveo/core/WOAndroidUtils.js');
}

var WOApplicationManager = App.Class(App.Controller.WOManager, {
	className	: 'WOApplicationManager',
	alias		: 'App',

	CONST		: {
		lastRunVersion  : 'WOApp.lastRunVersion',
		lastLaunched	: 'WOApp.lastLaunchTime',	// stored property, Time of last app launch/resume
		lastPaused		: 'WOApp.lastPausedTime',	// stored property, time of last app pause
		sessionTime		: 15 * 60 * 1000,			// 15mins in milliseconds
	},
	
    //--- Event Name constants -----
    EVENT		: {
		//--- responds to these events:
		doUpgrade		: 'App.doUpgrade',
		doRestart       : 'App.doRestart',
		doStart			: 'App.doStart',
		doRun			: 'App.doRun',
		DNDchange		: 'App.DNDchange',			// change in Do-Not-Disturb status
		
		pause			: 'pause',					// system event
		resume			: 'resume',					// system event
		//--- produces these events:
		ready         	: 'App.ready',				// app lauched or resumed, AND logged in
		run             : 'App.run',
		stop            : 'App.stop',
    },
    
	//--- Usage ACTIONs ------------
	USAGE		: {
		session			: 'session',
	},

	/**
	 * MUST provide necessary services for UpgradeManager to function
	 */
	initialize : function(param) {
		//---  Sub-class MUST call this method using self.base() before own behaviors		
		var self = this;
		self.base(param);

		self.isDND		= true;		// make sure no interruption pre-registration & pre-auth
		self.autoResetTime = 0;
		self.isLogin	= false;
		self.hasStarted	= false;
		
		self.lastRouteTime	= new Date();	// used for preventing multiple routes
		self.lastRoute		= '';
	},

	initManagers : function(param){
		var self = this;
		
		//--- load baseline (memory-resident) managers ---
		$.configMgr   = App.Helper.newController('ConfigManager');
		$.accountMgr  = App.Helper.newController('AccountController');
		$.syncMgr     = App.Helper.newController('SyncManager');
		$.cardLibMgr  = App.Helper.newController('WOCardLibManager');
		$.dataMgr	  = App.Helper.newController('DataManager');
		$.trackMgr    = App.Helper.newController('TrackManager');
		$.watch		  = App.Helper.newController('WatchDog');
        //--- ONLY hook up Event Listeners allowed to respond PRE-LOGIN ! ---
        //    Hook up rest of Event handlers POST-AUTH
	},

    /**
     * check if this is the very FIRST time app is run after INSTALL or UPGRADE
     * @return  {boolean}   TRUE when is first-run
     */
    isFirstRun : function(){
		var self = this;
		
		var lastRunVersion = Ti.App.Properties.getString(self.CONST.lastRunVersion, null);
		if(Ti.App.version == lastRunVersion) return false;
		return true;
    },
	
	getLastRunVersion : function(){
		var self = this;
		
        var lastRunVersion = Ti.App.Properties.getString(self.CONST.lastRunVersion, null);
        var type = '>=2.3';
        
        if(!lastRunVersion) {
        	//check is version 1.x, use old dbi, make sure be compatible
        	type = '1.x';
        	lastRunVersion = $.configMgr.getOld('version', type);
        	if(!lastRunVersion) lastRunVersion = '1.1';
        	if(lastRunVersion == 'new') {
        		type = '2.x';
        		lastRunVersion = $.configMgr.getOld('version', type);
        	}
        	if('new' == lastRunVersion) { //means >= 2.3.x, from 2.3.x, we changed account db, set version to 'new'
        		type = null;
        		lastRunVersion = null;
        	}
        }
        return {
        
        	version : lastRunVersion,
        	type	: type,
        }
	},
	
    //---- NO database-dependent operations allowed for methods ABOVE this point ----
     
	start : function(){
        var self = this;
		self.hasStarted = true;
	},
	
    run : function(){
    },

    stop : function(){
        var self = this;
        self.hasStarted = false;
    },

    /**
     * Called only ONCE when it is the very FIRST time app is run after INSTALL or UPGRADE
     */
    firstRun : function(param){
        // sub-class to implement app-specific behavior if necessary      
        var self = this;
        //init sync timestamp
       	for(var className in App.Config.SYNC_TIME_INI[App.DBSource]) {
        	if(!$.syncMgr.getLastSyncTimestamp(className)) $.syncMgr.setLastSyncTimestamp(className, App.Config.SYNC_TIME_INI[App.DBSource][className]);
        }
        //copy prepackaged images to images directory
        var prepackaged_imagesDir = Ti.Filesystem.getFile(Ti.Filesystem.resourcesDirectory, 'assets/images/prepackaged/'+App.DBSource);
		var imagesDir = Ti.Filesystem.getFile(App.Config.IMAGE_DIR);
		if(!imagesDir.exists()) imagesDir.createDirectory();
		//copy prepackaged images to image folder
		if (prepackaged_imagesDir.exists()) {
			var fileList = prepackaged_imagesDir.getDirectoryListing();
		    if (App.Config.IS_ANDROID) {
		    	//TODO [ZJ] current nothing todo.
		    } else {
			    for(var i=0; i<fileList.length; i++){
			    	var srcImageFile  = Ti.Filesystem.getFile(prepackaged_imagesDir.getNativePath(), fileList[i]);
			    	var desImageFile = Ti.Filesystem.getFile(App.Config.IMAGE_DIR+srcImageFile.getName());
			    	desImageFile.write(srcImageFile.read());
			    }
		    }
		}
		//resize some default image size for Android platform, ios no need.
		if(App.Config.IS_ANDROID) {
			var imageFactory = require('ti.imagefactory');
			var fileList = [
				'default_faceAndroid.png',
				'default_thumbAndroid.png',
			];
			var ui_imagesDir = Ti.Filesystem.getFile(Ti.Filesystem.resourcesDirectory, 'assets/images/ui');
			for(var i=0; i<fileList.length; i++){
				var srcImageFile  = Ti.Filesystem.getFile(ui_imagesDir.getNativePath(), fileList[i]);
				var desImageFile = Ti.Filesystem.getFile(App.Config.IMAGE_DIR+fileList[i]);
			    desImageFile.write(imageFactory.imageAsResized(srcImageFile.read(), { width: App.Config.CardBigImageW, height: App.Config.CardBigImageH })); 
			}
		}
		
		//patch for 2.3 upgrade issue.
		if('2.3' == Ti.App.Properties.getString(self.CONST.lastRunVersion, null)) {
			//if user don't have PerkdID card or card master, force reset all objects last sync time to 0.
			if(!$.dataMgr.get('WOCardMaster', '000001', false) || !$.dataMgr.getRelated('WOCardMaster', '000001', 'WOCard').length) {
				Ti.API.error('------------------------------resetSyncTimestamp for 2.3 upgrade issue------------------------------');
				$.syncMgr.resetSyncTimestamp();
			}
		}
		
        //check upgrade
        var upgradeMgr = self._loadUpgradeMgr();
        if(! App.Config.IS_ANDROID && upgradeMgr.needUpgrade()){
            $.app.route({
                url			: 'UpgradeController/upgrade',
                effect		: App.Const.WINDOW.effect.flip,
                data		: {
                	callback	: _firstRunDone,
                }
            });
		}else{
			_firstRunDone();
		}

		//-----------  Callback Handler  ------------
		function _firstRunDone(){
	        upgradeMgr = null;
	        
	        //Track upgrade usage.
	        var lastRunVersion = Ti.App.Properties.getString(self.CONST.lastRunVersion, null);
			if(lastRunVersion) $.trackMgr.track($.app, 'upgrade', {from:lastRunVersion, to:Ti.App.version});
			
	        Ti.App.Properties.setString(self.CONST.lastRunVersion, Ti.App.version);
	        App.Version  = Ti.App.version;
	        
	        //---  Start the App !  ---
	        self.start();
		}
    },

    /**
     * handler when app being switched to background
     */
    pause : function(param){
    },

    /**
     * handler when app resume to foreground
     */
    resume : function(param){
        //--- Get access to App arguments/parameters passed in via URL activation of app
		// var arguments = JSON.stringify(Ti.App.getArguments());
    },

    /**
     * Go to a View / Window
     * @param	{Object}	abstract object that defines the routing:
     * 			url:		controllerName / methodName
     * 			data:		to be passed on as parameter to the view controller
     * 			style:		transition effect when display the new view/window
     * 			animated:	boolean
     */
    route : function(param){
        var self = this;
        param.data = param.data || {};

        //---- invoke Controller's Action method which returns a WOWindow object  ----
		var route = self.parseRoute(param.url);
		
		//---  filter out multiple-clicks  ---
		var now = new Date();
		if(now - self.lastRouteTime < 1600){
			if(self.lastRoute == param.url){
				self.lastRouteTime = now;
				return null;
			}
		}
		self.lastRouteTime = now;
		self.lastRoute = param.url;

        var winObj = route.controller[route.action](param);
        if(winObj != null){				//---  may return null due to error, just ignore
			winObj.open(param);
			$.currentWindow = winObj;          	
        }     
        return winObj;
    },

	parseRoute : function(path){
		var self = this;

		var route = {};
        var parts = path.split('/');
        var controllerName  = parts[0];
        route.action = parts[1];
        
        //---- load the Controller object ----
        if(controllerName in App.ControllerAlias){
            // controller already exist, just use it
            route.controller = App.ControllerAlias[controllerName];
            Ti.API.debug(self.className+'::parseRoute - using existing controller: '+route.controller.className+'('+controllerName+')');
        } else {
            // instantiate a new one
            route.controller = App.Helper.newController(controllerName);
        }
        return route;
	},

/**
 * @param	dndStatus
 * @param	autoResetTime		(optional) auto set DND to false after milliseconds
 */   
    doNotDisturb : function(dndStatus, autoResetTime){
        var self = this;
        autoResetTime = autoResetTime || 0;
        
        if(dndStatus === self.isDND && autoResetTime == self.autoResetTime) return;
        
        if(!dndStatus && self.autoResetTime > 0 && !(autoResetTime + self.autoResetTime)) {
        	self.isDND = dndStatus;
        	self.autoResetTime = 0;
        	Ti.App.fireEvent(self.EVENT.DNDchange, { isDND: self.isDND });
        } else if(dndStatus || !self.autoResetTime) {
        	self.isDND = dndStatus;
        	self.autoResetTime = autoResetTime > 0 ? autoResetTime : 0;
	        if(self.isDND && autoResetTime > 0){
	        	setTimeout(function(evt){
	        		self.doNotDisturb(false, -1*autoResetTime);
	        	}, autoResetTime);
	        }
	        Ti.App.fireEvent(self.EVENT.DNDchange, { isDND: self.isDND });	
        }
    },
    
    isStarted : function(){
    	var self = this;
		return self.hasStarted;
    },

	canDisturb : function(){
		var self = this;
		return (!self.isDND);	
	},

    notify2 : function(param){
        var self = this;

        if(App.Config.IS_ANDROID && param.message) {
            //alert(param.message);
            //return;
        }

		if (App.Config.IS_ANDROID) {
	        var win = createAndroidWindow({
	            backgroundColor : 'transparent',
	            touchEnabled : false
	         });
		} else {
	        var win = Ti.UI.createWindow({
	            backgroundColor : 'transparent',
	            touchEnabled : false
	        });
	   	}

        var view = Ti.UI.createView({
            width : '300dp',
            height : '40dp',
            top : '-40dp',
            backgroundColor : '#79000000',
            borderRadius: App.Config.IS_ANDROID? 5*Titanium.Platform.displayCaps.dpi/160 : 5,
        });

        var msg = Ti.UI.createLabel({
            text : param.message,
            color : '#FFFFFF',
            height : '30dp',
            bottom : 0
        });
        view.add(msg);

        win.add(view);
        win.open();

        var showAnimate = Ti.UI.createAnimation({
            top : '-10dp',
            duration : 500
        });
        var hideAnimate = Ti.UI.createAnimation({
            top : '-40dp',
            duration : 500,
            delay : 2000
        });
        view.animate(showAnimate);

        showAnimate.addEventListener('complete', function() {
            view.animate(hideAnimate);
        });
        hideAnimate.addEventListener('complete', function() {
            win.close();
        });
    },
    
    notify : function(param){
        var message = param.message;
	
        var win = Ti.UI.createWindow({
            width:'200dp', height:'54dp', zIndex:1000,
            hiding:false, //touchEnabled:false
			layout: 'composite',
        });
         var view = Ti.UI.createView({
            width:'200dp', height:'54dp', zIndex:1000,
            borderRadius: App.Config.IS_ANDROID? 10*Titanium.Platform.displayCaps.dpi/160 : 10,
			backgroundColor:'#000000',
            opacity:0.8,//transform: t1
            hiding:false, //touchEnabled:false
			layout: 'composite', touchEnabled:false,
        });      
        var labelMessage = Ti.UI.createLabel({
        	zIndex:1000,
            color:'white', textAlign:'center', text:message, touchEnabled:false,
			font:{fontSize:'21sp', fontFamily:'Melbourne-Bold'},
			minimumFontSize:'15sp',
			height:Ti.UI.SIZE,
			left:'10dp', right:'10dp',
        });
        view.add(labelMessage);
        win.add(view);
        
        var hideAnimate = Ti.UI.createAnimation({opacity:0,  duration:200});
        
        var timer = setTimeout(function(e){
            win.animate(hideAnimate);
        }, 2000);
        
        var timer2 = setTimeout(function(){
            closeNotify();
        }, 3000);
        
        win.addEventListener('click', function(e){
            closeNotify();
        });
        
        function closeNotify(){
            if(win.hiding) return;
            clearTimeout(timer);
            clearTimeout(timer2);
            win.hiding = true;
            win.close();
        }
        win.open();
    },    //TODO [WX] move to style file

    //-----------------------  Protected Methods  ------------------------

    /**
     * @return  {object}    Upgrade Manager object
     */
    _loadUpgradeMgr : function(){
        // app-specific Application Controller MAY override this method to load own UpgradeMgr
        var self = this;
        return App.Helper.newController('UpgradeController');
    },

});

module.exports = WOApplicationManager;

//---   End of file: WOApplicationManager.js
