/**
 * Base Factory class for making Widgets, ie. mini-controller object with pre-built views
 * 
 * @author young@waveo.com
 */
Ti.include('/globals.js');

var WOFactory = App.Class({
	className : 'WOFactory',

	//--- All widgets produced by this factory will deal with these events:
	WIDGET_EVENT	: {
		//--- Handles these events:
		refresh		: 'WOFactory.refresh',
		setMode		: 'WOFactory.setMode',
		//--- Produces these events:
		click		: 'WOFactory.click',
    },
    	
	MODES	: ['display', 'edit'],

	init : function(param){
	},

	//---  Constructor  --------------- 
	initialize : function(param){
		var self = this;
		
		self.widgetParam = {};		// stores widget definition & parameters (eg. _view, etc)
	},

	make : function(definition, data){
		var self = this;

		self.widgetParam = self._getDefinition(definition, data);

		//--------  Construct Views  -------------------------
		self.widgetParam._view = self._createView();

		//--------  Create Widget from Prototype  ------------
		var _Widget_ = self._getWidgetPrototype();
		var widget = new _Widget_(self.widgetParam);
		
		//--------  Customize Behaviors  ---------------------
		self._attachProperties(widget);			// NOTE:  attached to actual widget, NOT prototype
		self._attachMethods(_Widget_);
		if(self.widgetParam._eventProxy){
			self._defineEvents(_Widget_);
		}
		
		//--------  Get Widget to Init  ----------------------
		if(self.widgetParam._eventProxy){
			widget.setupEventHandlers();
		}
		widget.setMode(widget.currentMode);

		return widget;
	},

	//---------------------------  Protected Methods  ---------------------------------
	_getDefinition : function(param, data){
		var self = this;
		
		//--- Extract Widget Definition ---
		var widgetParam = {};
		widgetParam.name		= param.name;
		widgetParam.style		= param.style || {};
		widgetParam.properties	= param.properties || {};
		widgetParam.data		= data || {};

		widgetParam.modes		= {};
		var modeNames = param.modes ? param.modes : self.MODES;
		for(var m in modeNames){
			widgetParam.modes[modeNames[m]] = {};
		}
		widgetParam.defaultMode = ('undefined' != typeof param.defaultMode) ? param.defaultMode : self.MODES[0];

		widgetParam.useRow = ('undefined' != typeof param.useRow) ? param.useRow : false;
		widgetParam._eventProxy = param.eventProxy || null;

		widgetParam._view = null;		// main view of widget
		widgetParam._row = null;		// encapsulate our _view within rowView when useRow == true
		
		return widgetParam;
	},

	_createView : function(){
		var self = this;

		var view = Ti.UI.createView(self.widgetParam.style);

		//---  Views for each display mode
		for(mode in self.widgetParam.modes){
			var modeView = self._drawView(mode);
			view.add(modeView);
		}
		
		//---  Wrap Main View as a Row if specified
		if(self.widgetParam.useRow){
			self.widgetParam._row = self._createRow();
			self.widgetParam._row.add(view);
		}
		return view;
	},
	
	/*
	_createView : function(){
		var self = this;

		var view = Ti.UI.createView(self.widgetParam.style);

		//---  Views for each display mode
		for(mode in self.widgetParam.modes){
			var modeView = self._drawView(mode);
			view.add(modeView);
		}
		
		//---  Wrap Main View as a Row if specified
		if(self.widgetParam.useRow){
			self.widgetParam._row = self._createRow();
			if (App.Config.IS_ANDROID) {
				for(mode in self.widgetParam.modes){
					var modeView1 = self._drawView(mode);
					modeView1.left = self.widgetParam.style.left;
					modeView1.right = self.widgetParam.style.right;
					self.widgetParam._row.add(modeView1);
				}
			} else {
				self.widgetParam._row.add(view);			
			}
		}
		return view;
	},
	*/

	_createRow : function(){
		var self = this;

		//---  Render the Row  ---
		var rowParam = {};
		if('undefined' != typeof self.widgetParam.properties.row){
			rowParam = self.widgetParam.properties.row.style;
		}
		rowParam.className = 'row';
//		rowParam.height = Ti.UI.SIZE;

		var row = Ti.UI.createTableViewRow(rowParam);
		return row;
	},
	
	_drawView : function(mode){
		var self = this;
		
		var method = '_draw_' + mode.toLowerCase() + 'View';
		if('function' != typeof self[method]){
			Ti.API.error(self.className+'::drawView - '+method+'() not implemented');
		}
		//--- add view of this mode to main field view ---
		var modeView = self[method]();
		modeView.visible = false;
		
		return modeView;
	},

	_getWidgetPrototype : function(){

		//----------------  Widget object constructor  -------------
		//  IMPORTANT:  MUST NOT put this as a class method! JClass will interfere

		return function F(param){
			var self = this;

			self.id		= param.id;
			self.name	= param.name;
			self.data	= param.data;
			self.modes	= param.modes;
			self.VIEW	= param._view;
			self._row	= param._row;
						
			self.EVENT = WOFactory.WIDGET_EVENT;
			self._eventProxy = param._eventProxy;
			
			self.currentMode = param.defaultMode;
		};
	},

	/**
	 * Add EVENTs that widget will deal with to eventProxy
	 * @param {Object} _Widget_
	 * @param {Object} moreEvents	optionally, add more custom events
	 */
	_defineEvents : function(_Widget_, moreEvents){
		var self = this;

		self.widgetParam._eventProxy.addEvents(self.WIDGET_EVENT);
		if(moreEvents){
			self.widgetParam._eventProxy.addEvents(moreEvents);
		}
	},

	_attachProperties : function(_Widget_){
		//--- instantiate any custom properties where necessary  ---
		// eg.  _Widget_.myProp = self.widgetParam.someProperty;
	},

	_attachMethods : function(_Widget_){
		var self = this;
		
		//---  attach object behaviors  ---
		_Widget_.prototype.getView	= self._getView_;
		_Widget_.prototype.getRow	= self._getRow_;
		_Widget_.prototype.setValue	= self._setValue_;
		_Widget_.prototype.getValue	= self._getValue_;
		_Widget_.prototype.blur		= self._blur_;
		_Widget_.prototype.dispose	= self._dispose_;

		_Widget_.prototype.refresh	= self._refresh_;
		_Widget_.prototype.setMode	= self._setMode_;
					
		if(self.widgetParam._eventProxy){
			//---  attach Event Handlers  -----
			_Widget_.prototype.doClick	= self._doClick_;

			//---  attach event handler installer method  ---
			_Widget_.prototype.setupEventHandlers	= self._setupEventHandlers_;
		}
	},
	
	
	//--------  View render Methods for each supported display Mode  ---------
	//  NOTE:  subclasses MUST implement

	_draw_displayView : function(){
		Ti.API.error(this.className+'::_draw_displayView - not implemented');
	},
	
	
	//----------  Template/Prototype Methods for Widget object  --------------

	_setupEventHandlers_ : function(){
		var self = this;
		
		//----------------- Android-conditional probably can combine. But must test ----------
		if (App.Config.IS_ANDROID) {
			self.VIEW.addEventListener('click', (function(that){
				return (function(evt){
					that.doClick(evt);
				});
			})(self));
			self._eventProxy.addEventListener(self._eventProxy.EVENT.refresh, (function(that){
				return (function(evt){
					that.refresh(evt.data);
				});
			})(self));
			self._eventProxy.addEventListener(self._eventProxy.EVENT.setMode, (function(that){
				return (function(evt){
					that.setMode(evt.mode);
				});
			})(self));
		} else {
			self.VIEW.addEventListener('click', function(evt){
				self.doClick(evt);
			});
			self._eventProxy.addEventListener(self._eventProxy.EVENT.refresh, function(evt){
				self.refresh(evt.data);
			});
			self._eventProxy.addEventListener(self._eventProxy.EVENT.setMode, function(evt){
				self.setMode(evt.mode);
			});			
		}


	},
		
	//--- subclasses MUST implement these methods  --------
	_setValue_ : function(data){
		Ti.API.error(this.className+'::_setValue_ - not implemented');
	},
	
	_getValue_ : function(){
		Ti.API.error(this.className+'::_getValue_ - not implemented');
	},
	
	_blur_ : function(){
	},
	
	//--- subclasses MAY override these methods  ----------
	_getView_ : function(){
		return this.VIEW;
	},

	_getRow_ : function(){
		return this._row;
	},

	_doClick_ : function(evt){
		var self = this;
		self._eventProxy.fireEvent(self._eventProxy.EVENT.click, evt);
	},
			
	_refresh_ : function(data){
		var self = this;
		
		//--- extract data property that matches my field name ---
		if('undefined' != typeof data[self.name]){
			self.setValue(data[self.name]);
		}
	},
		
	_setMode_ : function(mode){
		var self = this;

		for(var m in self.modes){
			self.modes[m].view.setVisible(m==mode);
		}		
//		self.modes[mode].view.setVisible(true);
		self.currentMode = mode;
	},
	
	_dispose_ : function(){
		var self = this;
		
		//--  we should remove event listeners ??  TODO
		self.VIEW = null;
	},
});

module.exports = WOFactory;

//---  End of file: WOFactory.js
